using System;
using System.Collections.Generic;
using System.Text;
using Microsoft.Win32.SafeHandles;
using System.Runtime.InteropServices;
using System.Threading;
using System.ComponentModel;
using log4net;
using log4net.Config;

namespace StradusUSBHID
{
    internal class StradusDevice
    {
        private static readonly ILog logger = LogManager.GetLogger(typeof(StradusDevice));


        private enum RequestTypes : int
        {
            SET_CMD_QRY = 0xA0,
            GET_RESPONSE_STATUS = 0xA1,
            GET_RESPONSE = 0xA2,
            SET_RESPONSE_RECEIVED = 0xA3,
        }
        //Constants for CreateFile() and other file I/O functions
        internal const short FILE_ATTRIBUTE_NORMAL = 0x80;
        internal const short INVALID_HANDLE_VALUE = -1;
        internal const uint GENERIC_READ = 0x80000000;
        internal const uint GENERIC_WRITE = 0x40000000;
        internal const uint CREATE_NEW = 1;
        internal const uint CREATE_ALWAYS = 2;
        internal const uint OPEN_EXISTING = 3;
        internal const uint FILE_SHARE_READ = 0x00000001;
        internal const uint FILE_SHARE_WRITE = 0x00000002;

        //Other constant definitions
        internal const uint DBT_DEVTYP_DEVICEINTERFACE = 0x05;
        internal const uint DEVICE_NOTIFY_WINDOW_HANDLE = 0x00;
        internal const uint ERROR_SUCCESS = 0x00;
        internal const uint ERROR_NO_MORE_ITEMS = 0x00000103;
        internal const uint SPDRP_HARDWAREID = 0x00000001;

        StradusHub myParentObj = null;

        Int32 _MyLaserIndex = 0;

        Byte[] OUTBuffer = new byte[65];	//Allocate a memory buffer equal to the OUT endpoint size + 1
        Byte[] INBuffer = new byte[65];		//Allocate a memory buffer equal to the IN endpoint size + 1

        //Byte[] MessageOut = new Byte[63];
        Byte[] RequestOut = new Byte[1];
        Byte[] MessageIn = new Byte[63];
        Byte[] RequestIn = new Byte[1];

        String requestOut;
        volatile String stringToSend;
        volatile byte sendNow = 0x00;
        byte STATE = 0x00;
        byte PSTATE = 0x00;
        private String myDevicePath;

        private SafeFileHandle WriteHandleToUSBDevice = null;
        private SafeFileHandle ReadHandleToUSBDevice = null;
        private BackgroundWorker bw;

        volatile bool readWriting = false;

        volatile String[] reply = new String[4];
        volatile String theReply;
        volatile bool hasReply = false;

        private Mutex rwl = new Mutex();
        private Mutex rwlReply = new Mutex();

        public Int32 MyLaserIndex
        {
            get { return _MyLaserIndex; }
            set { _MyLaserIndex = value; }
        }
    
        public unsafe StradusDevice(StradusHub myParent,String devicePath)
        {
            uint ErrorStatusWrite;
            uint ErrorStatusRead;

            myDevicePath = devicePath;
            myParentObj = myParent;
   
            if (myDevicePath != null)
            {
                //We now have the proper device path, and we can finally open read and write handles to the device.
                WriteHandleToUSBDevice = CreateFile(myDevicePath, GENERIC_WRITE, FILE_SHARE_READ | FILE_SHARE_WRITE, IntPtr.Zero, OPEN_EXISTING, 0, IntPtr.Zero);
                ErrorStatusWrite = (uint)Marshal.GetLastWin32Error();
                ReadHandleToUSBDevice = CreateFile(myDevicePath, GENERIC_READ, FILE_SHARE_READ | FILE_SHARE_WRITE, IntPtr.Zero, OPEN_EXISTING, 0, IntPtr.Zero);
                ErrorStatusRead = (uint)Marshal.GetLastWin32Error();

                if ((ErrorStatusWrite == ERROR_SUCCESS) && (ErrorStatusRead == ERROR_SUCCESS))
                {
                    open(myDevicePath);                   
                }//endif error check
                else
                {
                    if (ErrorStatusWrite == ERROR_SUCCESS)
                        WriteHandleToUSBDevice.Close();
                    if (ErrorStatusRead == ERROR_SUCCESS)
                        ReadHandleToUSBDevice.Close();
                }
            } //end if !=null
            XmlConfigurator.Configure(); //Required for to enable logging using log4net

        }
        //Accessor Methods&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&
        public String Path { get { return myDevicePath; } }
        //&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&

        #region DLL IMPORTS

        //Uses a handle (created with CreateFile()), and lets us write USB data to the device.
        [DllImport("kernel32.dll", SetLastError = true, CharSet = CharSet.Unicode)]
        static extern bool WriteFile(
            SafeFileHandle hFile,
            byte[] lpBuffer,
            uint nNumberOfBytesToWrite,
            ref uint lpNumberOfBytesWritten,
            IntPtr lpOverlapped);

        //Uses a handle (created with CreateFile()), and lets us read USB data from the device.
        [DllImport("kernel32.dll", SetLastError = true, CharSet = CharSet.Unicode)]
        static extern bool ReadFile(
            SafeFileHandle hFile,
            IntPtr lpBuffer,
            uint nNumberOfBytesToRead,
            ref uint lpNumberOfBytesRead,
            IntPtr lpOverlapped);

        //Takes in a device path and opens a handle to the device.
        [DllImport("kernel32.dll", SetLastError = true, CharSet = CharSet.Unicode)]
        static extern SafeFileHandle CreateFile(
            string lpFileName,
            uint dwDesiredAccess,
            uint dwShareMode,
            IntPtr lpSecurityAttributes,
            uint dwCreationDisposition,
            uint dwFlagsAndAttributes,
            IntPtr hTemplateFile);

        #endregion

        private void open(String myPath)
        {
            //Use path to get handles
            bw = new BackgroundWorker();
            bw.WorkerReportsProgress = true;
            bw.DoWork += new DoWorkEventHandler(bw_DoWork);
            readWriting = true;
            bw.RunWorkerAsync();
        }

        public void close()
        {
            //Stop the worker thread first then close the handles
            STATE = 0x06; //Don't do any transfers, just idle and wait for thread to terminate...
            readWriting = false;
            if(WriteHandleToUSBDevice!=null)
                WriteHandleToUSBDevice.Close();
            if(ReadHandleToUSBDevice!=null)
                ReadHandleToUSBDevice.Close();
        }

        private void SendRequest(Byte reqType, String message)
        {
            int p = 0;

            if (message != null)
            {
                RequestOut[0] = reqType;
                requestOut = bytes2hex(RequestOut);
                OUTBuffer[0] = 0x00;	//The first byte is the "Report ID" and does not get sent over the USB bus.  Always set = 0.
                OUTBuffer[1] = RequestOut[0]; //Control Byte, like vendor request
                Buffer.BlockCopy(Encoding.Default.GetBytes(message), 0, OUTBuffer, 2, message.Length); //src,dst

                //Initialize the rest of the 64-byte packet to "0xFF".  Binary '1' bits do not use as much power, and do not cause as much EMI
                //when they move across the USB cable.  USB traffic is "NRZI" encoded, where '1' bits do not cause the D+/D- signals to toggle states.
                //This initialization is not strictly necessary however.

                p = message.Length + 2;
                Buffer.BlockCopy(Encoding.Default.GetBytes(message), 0, OUTBuffer, 2, message.Length); //src,dst                               
                for (int i = p; i < 65; i++)
                    OUTBuffer[i] = 0xFF;
            }
        }

        public void sendString(String myString)
        {
            rwl.WaitOne();
            stringToSend = myString;
            rwl.ReleaseMutex();
            sendNow = 0x01;
            logger.Info("SendString: " + myString);
        }

        public String getReply()
        {
            String myReply;

            rwlReply.WaitOne();

            if (hasReply)
            {
                myReply = theReply;
                hasReply = false;
            }
            else
                myReply = "EMPTY";


            rwlReply.ReleaseMutex();
            logger.Info("Get Reply: " + myReply);
            return myReply;
        }


        #region USB READ BACKGROUND WORKER THREAD

        private void bw_DoWork(object sender, DoWorkEventArgs e)
        {
            uint BytesWritten = 0;
            uint BytesRead = 0;
            byte responsePending = 0x00;
            byte responseReceived = 0x00;
            byte armGetResponse = 0x00;
            Mutex myMute;

            while (true)
            {
                if (readWriting)	//Do not try to use the read/write handles unless the USB device is attached and ready
                {
                    if ((WriteHandleToUSBDevice != null) && (ReadHandleToUSBDevice != null))
                    {
                        Thread.Sleep(20);
                        try
                        {
                            if (ReadFileManagedBuffer(ReadHandleToUSBDevice, INBuffer, 65, ref BytesRead, IntPtr.Zero))	//Blocking function, unless an "overlapped" structure is used	
                            {
                                if (BytesRead > 0)
                                {
                                    //INBuffer[0] is the report ID, which we don't care about.
                                    //INBuffer[1] is the request type
                                    //INBuffer[2] to INBuffer[63] contain the data

                                    Buffer.BlockCopy(INBuffer, 1, RequestIn, 0, 1); //src,dst

                                    //Parse the fields of the packet received
                                    switch (RequestIn[0])
                                    {
                                        case (byte)RequestTypes.GET_RESPONSE_STATUS:
                                            responsePending = Buffer.GetByte(INBuffer, 2);
                                            //updateRXText("\r\nRx: " + requestIn + " GET_RESPONSE_STATUS:" + System.Convert.ToString(responsePending));
                                            break;
                                        case (byte)RequestTypes.GET_RESPONSE:
                                            if (armGetResponse == 0x01) //Avoid stale data
                                            {
                                                armGetResponse = 0x00;
                                                responseReceived = 0x01;
                                                Buffer.BlockCopy(INBuffer, 2, MessageIn, 0, 63); //src,dst    
                                            }
                                            //updateRXText("\r\nRx: " + requestIn + " GET_RESPONSE: " + Encoding.Default.GetString(MessageIn));
                                            break;
                                        default:
                                            break;
                                    }
                                    //System.Console.WriteLine("ReqIn: "+RequestIn[0].ToString());
                                }
                            }

                            //System.Console.WriteLine("STATE: "+STATE);
                            if (STATE != PSTATE)
                            {
                                logger.Info("STATE: " + STATE);
                                PSTATE = STATE;
                            }

                            //Finite State Machine
                            switch (STATE)
                            {
                                case 0x00: //Idle state
                                    //Poll this status continually in the idle state
                                    SendRequest((byte)RequestTypes.GET_RESPONSE_STATUS, String.Empty);
                                    WriteFile(WriteHandleToUSBDevice, OUTBuffer, 65, ref BytesWritten, IntPtr.Zero);
                                    if (responsePending == 0x01) //The firmware has indicated that there is a response ready to retrieve
                                    {
                                        //Flag is reset by firmware
                                        STATE = 0x02;
                                    }
                                    else
                                        if (sendNow == 0x01)
                                        {
                                            sendNow = 0x00;
                                            //Send a command now...
                                            STATE = 0x03;
                                        }  
                                    break;
                                case 0x01: //Set Response Received
                                    SendRequest((byte)RequestTypes.SET_RESPONSE_RECEIVED, String.Empty);
                                    WriteFile(WriteHandleToUSBDevice, OUTBuffer, 65, ref BytesWritten, IntPtr.Zero);
                                    if (myParentObj != null)
                                        System.Console.WriteLine("SET_RESPONSE_RECEIVED VR Sent...");
                                    STATE = 0x05;
                                    break;
                                case 0x02: //Get response now
                                    //Send request to get the response
                                    SendRequest((byte)RequestTypes.GET_RESPONSE, String.Empty);
                                    WriteFile(WriteHandleToUSBDevice, OUTBuffer, 65, ref BytesWritten, IntPtr.Zero);
                                    armGetResponse = 0x01;
                                    if(myParentObj!=null)
                                        System.Console.WriteLine("GET_RESPONSE VR Sent...");
                                    STATE = 0x04;
                                    break;
                                case 0x03: //Send a new command to laser
                                    rwl.WaitOne();
                                    SendRequest((byte)RequestTypes.SET_CMD_QRY, stringToSend);                                   
                                    WriteFile(WriteHandleToUSBDevice, OUTBuffer, 65, ref BytesWritten, IntPtr.Zero);
                                    rwl.ReleaseMutex();
                                    if (myParentObj != null)
                                        System.Console.WriteLine("SET_COMMAND VR Sent... " + stringToSend);                                  
                                    STATE = 0x00;
                                    break;
                                case 0x04: //WAIT STATE (Wait for the response to actually come back)
                                    if (responseReceived == 0x01)
                                    {
                                        responseReceived = 0x00;
                                        if (myParentObj != null)
                                            System.Console.WriteLine("Response from firmware is:" + Encoding.Default.GetString(MessageIn)); 
                                        STATE = 0x01;
                                    }
                                    break;
                                case 0x05: //WAIT STATE (Wait for GET_RESPONSE_STATUS to return 0)
                                    SendRequest((byte)RequestTypes.GET_RESPONSE_STATUS, String.Empty);
                                    WriteFile(WriteHandleToUSBDevice, OUTBuffer, 65, ref BytesWritten, IntPtr.Zero);
                                    if (responsePending == 0)
                                    {
                                        rwlReply.WaitOne();
                                        theReply = Encoding.Default.GetString(MessageIn);
                                        hasReply = true;
                                        rwlReply.ReleaseMutex();
                                        STATE = 0x00;
                                        System.Console.WriteLine("Command Transaction Complete");
                                    }
                                    break;
                                case 0x06:
                                    STATE = 0x06;
                                    break;
                            }
                            
                        }
                        catch
                        {
                            throw;
                            //Exceptions can occur during the read or write operations.  For example,
                            //exceptions may occur if for instance the USB device is physically unplugged
                            //from the host while the above read/write functions are executing.
                        }

                    }
                    else
                    {
                        Thread.Sleep(200);	//was 200, was 5ms Add a small delay.  Otherwise, this while(true) loop can execute very fast and cause 
                        //high CPU utilization, with no particular benefit to the application.
                    }



                } //end of while(true) loop
                //-------------------------------------------------------END CUT AND PASTE BLOCK-------------------------------------------------------------------------------------
                //-------------------------------------------------------------------------------------------------------------------------------------------------------------------
            }
        }
        #endregion

        //-------------------------------------------------------------------------------------------------------------------------------------------------------------------
        //-------------------------------------------------------BEGIN CUT AND PASTE BLOCK-----------------------------------------------------------------------------------

        //--------------------------------------------------------------------------------------------------------------------------
        //FUNCTION:	ReadFileManagedBuffer()
        //PURPOSE:	Wrapper function to call ReadFile()
        //
        //INPUT:	Uses managed versions of the same input parameters as ReadFile() uses.
        //
        //OUTPUT:	Returns boolean indicating if the function call was successful or not.
        //          Also returns data in the byte[] INBuffer, and the number of bytes read. 
        //
        //Notes:    Wrapper function used to call the ReadFile() function.  ReadFile() takes a pointer to an unmanaged buffer and deposits
        //          the bytes read into the buffer.  However, can't pass a pointer to a managed buffer directly to ReadFile().
        //          This ReadFileManagedBuffer() is a wrapper function to make it so application code can call ReadFile() easier
        //          by specifying a managed buffer.
        //--------------------------------------------------------------------------------------------------------------------------
        public unsafe bool ReadFileManagedBuffer(SafeFileHandle hFile, byte[] INBuffer, uint nNumberOfBytesToRead, ref uint lpNumberOfBytesRead, IntPtr lpOverlapped)
        {
            IntPtr pINBuffer = IntPtr.Zero;

            try
            {
                pINBuffer = Marshal.AllocHGlobal((int)nNumberOfBytesToRead);    //Allocate some unmanged RAM for the receive data buffer.

                if (ReadFile(hFile, pINBuffer, nNumberOfBytesToRead, ref lpNumberOfBytesRead, lpOverlapped))
                {
                    Marshal.Copy(pINBuffer, INBuffer, 0, (int)lpNumberOfBytesRead);    //Copy over the data from unmanged memory into the managed byte[] INBuffer
                    Marshal.FreeHGlobal(pINBuffer);
                    return true;
                }
                else
                {
                    Marshal.FreeHGlobal(pINBuffer);
                    return false;
                }
            }
            catch
            {
                if (pINBuffer != IntPtr.Zero)
                {
                    Marshal.FreeHGlobal(pINBuffer);
                }
                return false;
            }
        }

        public static string bytes2hex(byte[] bytes)
        {
            return bytes2hex(bytes, 0, bytes.Length);
        }

        public static string bytes2hex(byte[] bytes, bool prepend0x)
        {
            return bytes2hex(bytes, 0, bytes.Length, prepend0x);
        }

        public static string bytes2hex(byte[] bytes, int offset, int length)
        {
            return bytes2hex(bytes, offset, length, false);
        }

        public static string bytes2hex(byte[] bytes, int offset, int length, bool prepend0x)
        {
            if (bytes == null)
                return string.Empty;

            StringBuilder str = new StringBuilder();
            for (int i = offset; i < length; i++)
            {
                if (prepend0x)
                    str.Append(" 0x");
                str.Append(bytes[i].ToString("X2"));
            }
            return str.ToString();
        }
    }

    


}
